from flex.serializers.common import (
    HomogenousDictSerializer,
)

from rest_framework import serializers


class RoleSerializer(serializers.Serializer):
    """
    Test serializer to serve as the fields for the RolesSerializer
    """
    name = serializers.CharField()
    age = serializers.IntegerField()


class RolesSerializer(HomogenousDictSerializer):
    """
    Test serializer to test the functionality of the HomogenousDictSerializer
    """
    value_serializer_class = RoleSerializer


def test_autopopulates_fields_based_on_data():
    data = {
        'field_a': {'name': 'Field A', 'age': 10},
        'field_b': {'name': 'Field b', 'age': 20},
    }
    serializer = RolesSerializer(data=data)
    assert serializer.is_valid(), serializer.errors

    assert 'field_a' in serializer.fields
    assert 'field_b' in serializer.fields


class ParentSerializer(serializers.Serializer):
    roles = RolesSerializer()


def test_autopopulates_when_used_as_a_field():
    data = {
        'roles': {
            'field_a': {'name': 'Field A', 'age': 10},
            'field_b': {'name': 'Field b', 'age': 20},
        },
    }
    serializer = ParentSerializer(data=data)
    assert serializer.is_valid(), serializer.errors

    field = serializer.fields['roles']
    assert 'field_a' in field.fields
    assert 'field_b' in field.fields

    assert serializer.is_valid(), serializer.errors
    try:
        ret = serializer.validated_data
    except AttributeError:
        ret = serializer.object

    roles = ret.get('roles', {})
    assert 'field_a' in roles
    assert 'field_b' in roles



from flex.serializers.core import ParameterSerializer
from flex.error_messages import MESSAGES

from tests.utils import assert_error_message_equal


def test_unknown_parameter_reference():
    serializer = ParameterSerializer(
        data=['SomeReference'],
        context={},
        many=True,
    )

    assert not serializer.is_valid()
    assert_error_message_equal(
        serializer.errors[0][0],
        MESSAGES['unknown_reference']['parameter'],
    )


def test_valid_parameter_reference():
    serializer = ParameterSerializer(
        data=['SomeReference'],
        context={
            'parameters': {'SomeReference': {}},
        },
        many=True,
    )

    assert serializer.is_valid()



import collections
import re
import six


def check_if_error_message_equal(formatted_msg, unformatted_msg):
    """
    Helper assertion for testing that a formatted error message matches the
    expected unformatted version of that error.
    """
    # Replace all `{}` style substitutions with `.*` so that we can run a regex
    # on the overall message, ignoring any pieces that would have been
    # dynamically inserted.
    if not isinstance(formatted_msg, six.string_types):
        raise ValueError(
            "formatted_msg must be a string: got `{0}`".format(
                repr(formatted_msg)),
        )
    if not isinstance(unformatted_msg, six.string_types):
        raise ValueError(
            "unformatted_msg must be a string: got `{0}`".format(
                repr(unformatted_msg)),
        )
    # replace any string formatters
    pattern = re.sub('\{.*\}', '.*', unformatted_msg)
    # replace any parenthesis
    pattern = re.sub('\(', '\(', pattern)
    pattern = re.sub('\)', '\)', pattern)

    return bool(re.compile(pattern).search(formatted_msg))


def assert_error_message_equal(formatted_msg, unformatted_msg):
    if not check_if_error_message_equal(formatted_msg, unformatted_msg):
        raise AssertionError(
            "`{0}` not found in `{1}`".format(
                formatted_msg, unformatted_msg,
            )
        )


def _find_message_in_errors(message, errors, namespace=''):
    results = []

    if isinstance(errors, six.string_types):
        if check_if_error_message_equal(errors, message):
            yield namespace
    elif isinstance(errors, collections.Mapping):
        for key, error in errors.items():
            for match in _find_message_in_errors(
                message,
                error,
                '.'.join((namespace, key)).strip('.'),
            ):
                yield match
    elif isinstance(errors, list):
        for index, error in enumerate(errors):
            for match in _find_message_in_errors(
                message,
                error,
                '.'.join((namespace, six.text_type(index))).strip('.'),
            ):
                yield match
    else:
        raise ValueError("Unsupported type")


def find_message_in_errors(*args, **kwargs):
    paths = tuple(_find_message_in_errors(*args, **kwargs))
    return paths


def _find_matching_paths(target_path, paths):
    target_parts = tuple(reversed(target_path.split('.')))
    for message_path in paths:
        message_parts = message_path.split('.')
        for target_part in target_parts:
            while message_parts:
                message_part = message_parts.pop()
                if message_part == target_part:
                    break
            else:
                break
        else:
            yield message_path


def find_matching_paths(*args, **kwargs):
    paths = tuple(_find_matching_paths(*args, **kwargs))
    return paths


def assert_message_in_errors(message, errors, target_path=None):
    paths = tuple(_find_message_in_errors(message, errors))
    if not paths:
        raise AssertionError("Message: `{0}` not found in errors: `{1}`".format(
            message, errors,
        ))

    if target_path is not None:
        if not tuple(_find_matching_paths(target_path, paths)):
            raise AssertionError("No paths matched `{0}`.  Tried `{1}`".format(
                target_path, paths,
            ))


def _enumerate_error_paths(errors, namespace=''):
    if isinstance(errors, six.string_types):
        yield namespace
    elif isinstance(errors, collections.Mapping):
        for key, value in errors.items():
            for path in _enumerate_error_paths(
                value,
                '.'.join((namespace, key)).strip('.'),
            ):
                yield path
    elif isinstance(errors, list):
        for index, error in enumerate(errors):
            for path in _enumerate_error_paths(
                error,
                '.'.join((namespace, six.text_type(index))).strip('.'),
            ):
                yield path
    else:
        raise ValueError("Unsupported Type")


def enumerate_error_paths(*args, **kwargs):
    return tuple(_enumerate_error_paths(*args, **kwargs))


def assert_path_not_in_errors(path, errors):
    assert not find_matching_paths(path, errors)


def generate_validator_from_schema(schema, **kwargs):
    from flex.serializers.core import SchemaSerializer

    serializer = SchemaSerializer(data=schema, **kwargs)
    assert serializer.is_valid(), serializer.errors

    validator = serializer.save()
    return validator



import pytest

from flex.error_messages import MESSAGES
from flex.utils import is_non_string_iterable
from flex.serializers.fields import (
    MaybeListCharField,
    SecurityRequirementReferenceField,
)
from flex.exceptions import ValidationError

from rest_framework import serializers

from tests.utils import assert_error_message_equal


class TestSerializer(serializers.Serializer):
    test_field = MaybeListCharField()

def get_validated_data(serializer):
    assert serializer.is_valid(), serializer.errors

    try:
        return serializer.validated_data
    except AttributeError:
        return serializer.object


#
# MaybeListCharField tests
#
def test_maybe_list_char_field_accepts_strings():
    data = {'test_field': 'a-string'}
    serializer = TestSerializer(data=data)
    actual = get_validated_data(serializer)
    assert actual['test_field'] == 'a-string'


def test_maybe_list_char_field_accepts_lists():
    data = {'test_field': ['a-string', 'another-string']}
    serializer = TestSerializer(data=data)
    actual = get_validated_data(serializer)
    assert actual['test_field'] == ['a-string', 'another-string']


def test_maybe_list_char_field_runs_validators_on_singular_strings():
    def validator(value):
        if is_non_string_iterable(value):
            if not all([v.startswith('bar') for v in value]):
                raise ValidationError('error')
        else:
            if not value.startswith('bar'):
                raise ValidationError('error')

    class WithValidatorSerializer(serializers.Serializer):
        test_field = MaybeListCharField(validators=[validator])

    data = {'test_field': 'not-bar'}
    serializer = WithValidatorSerializer(data=data)
    with pytest.raises(AssertionError):
        get_validated_data(serializer)


def test_maybe_list_char_field_runs_validators_on_lists():
    def validator(value):
        if is_non_string_iterable(value):
            if not all([v.startswith('bar') for v in value]):
                raise ValidationError('error')
        else:
            if not value.startswith('bar'):
                raise ValidationError('error')

    class WithValidatorSerializer(serializers.Serializer):
        test_field = MaybeListCharField(validators=[validator])

    data = {'test_field': ['a-string', 'another-string']}
    serializer = WithValidatorSerializer(data=data)
    with pytest.raises(AssertionError):
        get_validated_data(serializer)


#
# SecurityRequirementReferenceField tests
#
def test_invalid_with_unknown_reference():
    class TestSerializer(serializers.Serializer):
        foo = SecurityRequirementReferenceField()

    serializer = TestSerializer(
        data={'foo': 'SomeReference'},
        context={
            'securityDefinitions': {},
        },
    )

    assert not serializer.is_valid()
    assert 'foo' in serializer.errors
    assert_error_message_equal(
        serializer.errors['foo'][0],
        MESSAGES['unknown_reference']['security'],
    )


def test_valid_with_known_reference():
    class TestSerializer(serializers.Serializer):
        foo = SecurityRequirementReferenceField()

    serializer = TestSerializer(
        data={'foo': 'SomeReference'},
        context={
            'securityDefinitions': {'SomeReference': {}},
        },
    )

    assert serializer.is_valid(), serializer.errors



from tests.utils import (
    check_if_error_message_equal,
    find_message_in_errors,
    find_matching_paths,
)


def test_foo():
    formatted_msg = u'Unknown type: unknown-type'
    unformatted_msg = u'Unknown type: {0}'

    assert check_if_error_message_equal(formatted_msg, unformatted_msg)


#
# find_message_in_errors
#
def test_matched_string_returns_namespace_path():
    """
    If a string is passed in as errors, the namespace should be returned if the
    string is a match.
    """
    errors = 'foo'
    path = find_message_in_errors('foo', errors, namespace='namespace')
    assert set(path) == set(['namespace'])


def test_non_matching_string_returns_empty():
    """
    If a string is passed in as errors, the namespace should be returned if the
    string is a match.
    """
    errors = 'foo'
    path = find_message_in_errors('not-a-match', errors)
    assert set(path) == set([])


def test_list_matches_return_element_index():
    errors = ['bar', 'foo']
    path = find_message_in_errors('foo', errors)
    assert set(path) == set(['1'])


def test_dictionary_matches_return_element_key():
    errors = {
        'bar': 'foo',
    }
    path = find_message_in_errors('foo', errors)
    assert set(path) == set(['bar'])


def test_nested_matches_return_full_namespace():
    errors = {
        'bar': ['foo'],
    }
    path = find_message_in_errors('foo', errors)
    assert set(path) == set(['bar.0'])


def test_a_complex_nested_match():
    errors = {
        'bar': [{'baz': 'foo'}],
    }
    path = find_message_in_errors('foo', errors)
    assert set(path) == set(['bar.0.baz'])


def test_multiple_matches_return_all_namespaces():
    errors = {
        'bar': [{'baz': 'foo'}],
        'bop': 'foo',
    }
    path = find_message_in_errors('foo', errors)
    assert set(path) == set(['bop', 'bar.0.baz'])


#
# find_message_paths tests
#
def test_simple_path_match():
    paths = ['foo']
    assert find_matching_paths('foo', paths)


def test_exact_path_match():
    paths = ['foo.bar']
    assert find_matching_paths('foo.bar', paths)


def test_missing_at_beginning_match():
    paths = ['0.foo.bar']
    assert find_matching_paths('foo.bar', paths)


def test_missing_at_middle_match():
    paths = ['foo.0.bar']
    assert find_matching_paths('foo.bar', paths)


def test_missing_at_end_match():
    paths = ['foo.bar.0']
    assert find_matching_paths('foo.bar', paths)


def test_simple_mismatch():
    paths = ['bar.foo']
    assert not find_matching_paths('foo.bar', paths)


def test_other_mismatch():
    paths = ['foo.0.bar']
    assert not find_matching_paths('foo.0.0.bar', paths)
